# authentic2 - versatile identity manager
# Copyright (C) 2010-2019 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.shortcuts import render
from django.utils.translation import ugettext as _, ugettext_lazy

from authentic2.a2_rbac.models import OrganizationalUnit as OU, Role
from authentic2.custom_user.models import User
from . import views, app_settings, utils, constants, models
from .forms import authentication as authentication_forms


class LoginPasswordAuthenticator(object):
    submit_name = 'login-password-submit'

    def enabled(self):
        return app_settings.A2_AUTH_PASSWORD_ENABLE

    def name(self):
        return ugettext_lazy('Password')

    def id(self):
        return 'password'

    def login(self, request, *args, **kwargs):
        service_slug = request.GET.get(constants.SERVICE_FIELD_NAME)
        context = kwargs.get('context', {})
        is_post = request.method == 'POST' and self.submit_name in request.POST
        data = request.POST if is_post else None
        initial = {}
        preferred_ous = []

        # Special handling when the form contains an OU selector
        if app_settings.A2_LOGIN_FORM_OU_SELECTOR:
            default_ou = None
            # for the special case of services open to only one OU, pre-select it
            if service_slug:
                roles = Role.objects.filter(allowed_services__slug=service_slug).children()
                ous = OU.objects.filter(id__in=User.objects.filter(roles__in=roles).values_list('ou_id', flat=True))
                if len(ous) == 1:
                    default_ou = ous[0]
            preferred_ous = utils.get_remember_cookie(request, 'preferred-ous')
            if preferred_ous:
                preferred_ous = OU.objects.filter(pk__in=preferred_ous)
            if not default_ou and preferred_ous:
                default_ou = preferred_ous[0]
            if default_ou:
                initial['ou'] = default_ou

        form = authentication_forms.AuthenticationForm(
            request=request,
            data=data,
            initial=initial,
            preferred_ous=preferred_ous)
        if app_settings.A2_ACCEPT_EMAIL_AUTHENTICATION:
            form.fields['username'].label = _('Username or email')
        if app_settings.A2_USERNAME_LABEL:
            form.fields['username'].label = app_settings.A2_USERNAME_LABEL
        is_secure = request.is_secure
        context['submit_name'] = self.submit_name
        if is_post:
            utils.csrf_token_check(request, form)
            if form.is_valid():
                if is_secure:
                    how = 'password-on-https'
                else:
                    how = 'password'
                if form.cleaned_data.get('remember_me'):
                    request.session['remember_me'] = True
                    request.session.set_expiry(app_settings.A2_USER_REMEMBER_ME)
                response = utils.login(request, form.get_user(), how,
                                       service_slug=service_slug)
                if 'ou' in form.fields:
                    utils.prepend_remember_cookie(request, response, 'preferred-ous', form.cleaned_data['ou'].pk)
                return response
        context['form'] = form
        return render(request, 'authentic2/login_password_form.html', context)

    def profile(self, request, *args, **kwargs):
        return views.login_password_profile(request, *args, **kwargs)

    def registration(self, request, *args, **kwargs):
        context = kwargs.get('context', {})
        return render(request, 'authentic2/login_password_registration_form.html', context)
